@@ -1,29 +1,32 @@
 from __future__ import annotations
 
+import datetime
 import logging
-import math
 import multiprocessing
 import os
+import os.path
 import random
 import tempfile
 import threading
 import time
 import timeit
 from multiprocessing.managers import SyncManager
-from typing import Type, TYPE_CHECKING, Tuple, Dict
+from typing import Type, TYPE_CHECKING, Tuple, Dict, Optional, Union
 
+import joblib
 from ConfigSpace.configuration_space import ConfigurationSpace
-from sklearn.pipeline import Pipeline
 
 from dswizard.core.base_structure_generator import BaseStructureGenerator
 from dswizard.core.config_cache import ConfigCache
+from dswizard.core.constants import MODEL_DIR
 from dswizard.core.dispatcher import Dispatcher
 from dswizard.core.ensemble import EnsembleBuilder
-from dswizard.core.logger import JsonResultLogger
-from dswizard.core.model import StructureJob, Dataset, EvaluationJob, CandidateStructure, CandidateId
+from dswizard.core.logger import ResultLogger
+from dswizard.core.model import StructureJob, Dataset, EvaluationJob, CandidateStructure, CandidateId, MetaInformation
+from dswizard.core.renderer import NotebookRenderer
 from dswizard.core.runhistory import RunHistory
-from dswizard.optimizers.bandit_learners import HyperbandLearner
-from dswizard.optimizers.config_generators import RandomSampling
+from dswizard.optimizers.bandit_learners import PseudoBandit
+from dswizard.optimizers.config_generators import Hyperopt
 from dswizard.optimizers.structure_generators.mcts import MCTS
 from dswizard.pipeline.voting_ensemble import PrefitVotingClassifier
 from dswizard.workers import SklearnWorker
@@ -32,6 +35,7 @@ if TYPE_CHECKING:
     from dswizard.core.base_bandit_learner import BanditLearner
     from dswizard.core.base_config_generator import BaseConfigGenerator
     from dswizard.core.worker import Worker
+    from dswizard.pipeline.pipeline import FlexiblePipeline
 
 
 class Master:
@@ -40,23 +44,23 @@ class Master:
                  working_directory: str = '.',
                  model: str = None,
                  logger: logging.Logger = None,
-                 result_logger: JsonResultLogger = None,
+                 result_logger: ResultLogger = None,
 
                  wallclock_limit: int = 60,
                  cutoff: int = None,
                  structure_cutoff_factor: float = 2.,
-                 pre_sample: bool = True,
+                 pre_sample: bool = False,
 
                  n_workers: int = 1,
                  worker_class: Type[Worker] = SklearnWorker,
 
-                 config_generator_class: Type[BaseConfigGenerator] = RandomSampling,
+                 config_generator_class: Type[BaseConfigGenerator] = Hyperopt,
                  config_generator_kwargs: dict = None,
 
                  structure_generator_class: Type[BaseStructureGenerator] = MCTS,
                  structure_generator_kwargs: dict = None,
 
-                 bandit_learner_class: Type[BanditLearner] = HyperbandLearner,
+                 bandit_learner_class: Type[BanditLearner] = PseudoBandit,
                  bandit_learner_kwargs: dict = None
                  ):
         """
@@ -76,9 +80,9 @@ class Master:
             structure_generator_kwargs = {}
 
         self.working_directory = working_directory
-        os.makedirs(self.working_directory, exist_ok=True)
+        self.temp_dir = tempfile.TemporaryDirectory()
         if 'working_directory' not in config_generator_kwargs:
-            config_generator_kwargs['working_directory'] = self.working_directory
+            config_generator_kwargs['working_directory'] = self.temp_dir.name
 
         if logger is None:
             self.logger = logging.getLogger('Master')
@@ -86,10 +90,12 @@ class Master:
             self.logger = logger
 
         if result_logger is None:
-            result_logger = JsonResultLogger(self.working_directory, overwrite=True)
+            result_logger = ResultLogger(self.working_directory, self.temp_dir.name)
         self.result_logger = result_logger
         self.jobs = []
-        self.meta_data = {}
+
+        # noinspection PyTypeChecker
+        self.meta_information: MetaInformation = None
 
         self.ds = ds
         self.ds.cutoff = cutoff
@@ -99,30 +105,46 @@ class Master:
         self.pre_sample = pre_sample
         self.abort = False
 
+        self.n_structures = 0
+
         # condition to synchronize the job_callback and the queue
         self.thread_cond = threading.Condition()
         self.incomplete_structures: Dict[CandidateId, Tuple[CandidateStructure, int, int]] = {}
 
-        SyncManager.register('StructureGenerator', structure_generator_class)
-        SyncManager.register('ConfigCache', ConfigCache)
-
-        self.mgr = multiprocessing.Manager()
-        # noinspection PyUnresolvedReferences
-        self.cfg_cache: ConfigCache = self.mgr.ConfigCache(clazz=config_generator_class,
-                                                           init_kwargs=config_generator_kwargs,
-                                                           model=model)
-        # noinspection PyUnresolvedReferences
-        self.structure_generator: BaseStructureGenerator = self.mgr.StructureGenerator(cfg_cache=self.cfg_cache,
-                                                                                       cutoff=self.cutoff,
-                                                                                       workdir=self.working_directory,
-                                                                                       model=model,
-                                                                                       wallclock_limit=wallclock_limit,
-                                                                                       **structure_generator_kwargs)
-
         if n_workers < 1:
-            raise ValueError('Expected at least 1 worker, given {}'.format(n_workers))
+            raise ValueError(f'Expected at least 1 worker, given {n_workers}')
+        elif n_workers == 1 and cutoff <= 0:
+            self.mgr: Optional[multiprocessing.Manager] = None
+            self.cfg_cache: ConfigCache = ConfigCache(
+                clazz=config_generator_class,
+                init_kwargs=config_generator_kwargs,
+                model=model)
+            self.structure_generator: BaseStructureGenerator = structure_generator_class(
+                cfg_cache=self.cfg_cache,
+                cutoff=self.cutoff,
+                workdir=self.working_directory,
+                model=model,
+                wallclock_limit=wallclock_limit,
+                **structure_generator_kwargs)
+        else:
+            SyncManager.register('StructureGenerator', structure_generator_class)
+            SyncManager.register('ConfigCache', ConfigCache)
+            self.mgr: Optional[multiprocessing.Manager] = multiprocessing.Manager()
+            # noinspection PyUnresolvedReferences
+            self.cfg_cache: ConfigCache = self.mgr.ConfigCache(
+                clazz=config_generator_class,
+                init_kwargs=config_generator_kwargs,
+                model=model)
+            # noinspection PyUnresolvedReferences
+            self.structure_generator: BaseStructureGenerator = self.mgr.StructureGenerator(
+                cfg_cache=self.cfg_cache,
+                cutoff=self.cutoff,
+                workdir=self.working_directory,
+                model=model,
+                wallclock_limit=wallclock_limit,
+                **structure_generator_kwargs)
+
         self.workers = []
-        self.temp_dir = tempfile.TemporaryDirectory()
         for i in range(n_workers):
             worker = worker_class(wid=str(i), cfg_cache=self.cfg_cache, workdir=self.temp_dir.name)
             self.workers.append(worker)
@@ -136,24 +158,39 @@ class Master:
         time.sleep(1)
         self.structure_generator.shutdown()
         self.dispatcher.shutdown()
-        self.mgr.shutdown()
+        if self.mgr is not None:
+            self.mgr.shutdown()
 
     def cleanup(self):
         self.temp_dir.cleanup()
 
-    def optimize(self) -> Tuple[Pipeline, RunHistory]:
+    def optimize(self, fit: bool = True,
+                 ensemble: bool = True,
+                 render: bool = True,
+                 store: bool = True) -> Union[Tuple[FlexiblePipeline, RunHistory],
+                                              Tuple[FlexiblePipeline, RunHistory, PrefitVotingClassifier]]:
         """
         run optimization
         :return:
         """
 
         start = timeit.default_timer()
-        self.meta_data['start'] = start
-        self.logger.info('starting run at {}. Configuration:\n'
-                         '\twallclock_limit: {}\n'
-                         '\tcutoff: {}\n'
-                         '\tpre_sample: {}'.format(time.strftime('%Y-%m-%dT%H:%M:%S%z', time.localtime(time.time())),
-                                                   self.wallclock_limit, self.cutoff, self.pre_sample))
+        start_time = datetime.datetime.now()
+
+        data_file = os.path.join(self.working_directory, 'dataset.pkl')
+        self.ds.store(data_file)
+
+        self.meta_information = MetaInformation(start_time=time.time(), metric=self.ds.metric,
+                                                openml_task=self.ds.task, openml_fold=self.ds.fold,
+                                                data_file=os.path.abspath(data_file),
+                                                config={
+                                                    'cutoff': self.cutoff,
+                                                    'wallclock_limit': self.wallclock_limit,
+                                                })
+        self.logger.info(f'starting run at {start_time:%Y-%m-%d %H:%M:%S}. Configuration:\n'
+                         f'\twallclock_limit: {self.wallclock_limit}\n'
+                         f'\tcutoff: {self.cutoff}\n'
+                         f'\tpre_sample: {self.pre_sample}')
         for worker in self.workers:
             worker.start_time = start
         deadline = start + self.wallclock_limit
@@ -182,6 +219,8 @@ class Master:
                     self.logger.info('Aborting optimization')
                     self.dispatcher.finish_work(self.cutoff)
                     return True
+                if self.n_structures > 200:
+                    return True
 
                 job = None
                 with self.thread_cond:
@@ -194,6 +233,7 @@ class Master:
                         config_id = candidate.cid.with_config(len(candidate.results) + running)
                         if self.pre_sample:
                             config, cfg_key = self.cfg_cache.sample_configuration(
+                                cid=config_id,
                                 configspace=candidate.pipeline.configuration_space,
                                 mf=self.ds.meta_features)
                             cfg_keys = [cfg_key]
@@ -213,10 +253,9 @@ class Master:
                         try:
                             candidate = next(it)
                             if candidate is None:
-                                self.logger.debug(
-                                    'Waiting for next job to finish. Currently {} running, {} outstanding'.format(
-                                        len(self.dispatcher.running_jobs),
-                                        self.bandit_learner.iterations[-1].num_running))
+                                self.logger.debug(f'Waiting for next job to finish. '
+                                                  f'Currently {len(self.dispatcher.running_jobs)} running, '
+                                                  f'{self.bandit_learner.iterations[-1].num_running} outstanding')
                                 # Safety-net to prevent infinite waiting
                                 # noinspection PyTypeChecker
                                 self.thread_cond.wait(max(self.cutoff, deadline - timeit.default_timer()))
@@ -242,7 +281,7 @@ class Master:
         #   incumbent, loss = bandit_learners.optimize(configspace, structure)
         #   Update score of selected structure with loss
 
-        # Main hyperparamter optimization logic
+        # Main hyperparameter optimization logic
         timeout = False
         repetition = 0
         offset = 0
@@ -250,7 +289,7 @@ class Master:
             while not timeout:
                 # noinspection PyTypeChecker
                 self.dispatcher.finish_work(max(self.cutoff, deadline - timeit.default_timer()))
-                self.logger.info('Starting repetition {}'.format(repetition))
+                self.logger.info(f'Starting repetition {repetition}')
                 self.bandit_learner.reset(offset)
                 timeout = _optimize()
                 repetition += 1
@@ -258,21 +297,45 @@ class Master:
         except KeyboardInterrupt:
             self.logger.info('Aborting optimization due to user interrupt')
         finally:
+            structure_explanations = self.structure_generator.explain()
+            config_explanations = self.cfg_cache.explain()
             self.shutdown()
 
-        self.meta_data['end'] = time.time()
-        self.logger.info('Finished run after {} seconds'.format(math.ceil(timeit.default_timer() - start)))
+        self.logger.info(f'Finished run after {(datetime.datetime.now() - start_time).seconds} seconds')
 
         iterations = self.result_logger.load()
         # noinspection PyAttributeOutsideInit
-        self.rh_ = RunHistory(iterations, {**self.meta_data, **self.bandit_learner.meta_data})
+        self.rh_ = RunHistory.create(iterations, self.meta_information, self.bandit_learner.meta_data,
+                                     os.path.join(self.working_directory, MODEL_DIR),
+                                     structure_explanations, config_explanations)
+        self.result_logger.log_run_history(self.rh_, str(self.meta_information.openml_task))
 
         pipeline, _ = self.rh_.get_incumbent()
-        return pipeline, self.rh_
 
-    def build_ensemble(self, ds: Dataset = None) -> PrefitVotingClassifier:
-        ensemble = EnsembleBuilder(self.temp_dir.name, self.result_logger.structure_fn)
-        return ensemble.fit(self.ds if ds is None else ds, self.rh_).get_ensemble()
+        if fit:
+            pipeline.fit(self.ds.X, self.ds.y)
+        if render and fit:
+            self.render(pipeline)
+        if store and fit:
+            joblib.dump(pipeline, os.path.join(self.working_directory, 'incumbent.pkl'))
+        if ensemble:
+            ensemble = self.build_ensemble(store=store)
+            return pipeline, self.rh_, ensemble
+        else:
+            return pipeline, self.rh_
+
+    def build_ensemble(self, ds: Dataset = None, store: bool = True) -> PrefitVotingClassifier:
+        builder = EnsembleBuilder(self.working_directory, self.result_logger.structure_fn)
+        ensemble = builder.fit(self.ds if ds is None else ds).get_ensemble()
+        if store:
+            self.result_logger.log_ensemble(ensemble, str(self.meta_information.openml_task))
+        return ensemble
+
+    def render(self, incumbent: FlexiblePipeline, ds: Dataset = None):
+        renderer = NotebookRenderer()
+        renderer.render(incumbent,
+                        self.ds if ds is None else ds,
+                        os.path.join(self.working_directory, 'incumbent.ipynb'))
 
     def _evaluation_callback(self, job: EvaluationJob) -> None:
         """
@@ -281,17 +344,17 @@ class Master:
         :param job: Finished Job
         :return:
         """
-        self.logger.debug('Evaluation callback {}'.format(job.cid))
+        self.logger.debug(f'Evaluation callback {job.cid}')
         with self.thread_cond:
             try:
                 if job.config is None:
                     self.logger.error(
-                        'Encountered job without a configuration: {}. Using empty config as fallback'.format(job.cid))
-                    job.config = ConfigurationSpace().get_default_configuration()
-
-                if self.result_logger is not None:
-                    self.result_logger.log_evaluated_config(job.cid, job.result)
+                        f'Encountered job without a configuration: {job.cid}. Using empty config as fallback')
+                    config = ConfigurationSpace().get_default_configuration()
+                    config.origin = 'Default'
+                    job.config = config
 
+                self.result_logger.log_evaluated_config(job.cs, job.result)
                 cs = self.bandit_learner.register_result(job.cs, job.result)
                 self.structure_generator.register_result(job.cs, job.result)
                 self.cfg_cache.register_result(job)
@@ -303,17 +366,16 @@ class Master:
             except KeyboardInterrupt:
                 raise
             except (BrokenPipeError, EOFError) as ex:
-                self.logger.fatal('Lost connection to SyncManager probably due to OOM. Aborting...: {}'.format(ex),
+                self.logger.fatal(f'Lost connection to SyncManager probably due to OOM. Aborting...: {ex}',
                                   exc_info=True)
                 self.abort = True
             except Exception as ex:
-                self.logger.fatal('Encountered unhandled exception {}. This should never happen!'.format(ex),
-                                  exc_info=True)
+                self.logger.fatal(f'Encountered unhandled exception {ex}. This should never happen!', exc_info=True)
             finally:
                 self.thread_cond.notify_all()
 
     def _structure_callback(self, cs: CandidateStructure):
-        self.logger.debug('Structure callback {}'.format(cs.cid))
+        self.logger.debug(f'Structure callback {cs.cid}')
         with self.thread_cond:
             try:
                 if cs.is_proxy():
@@ -328,19 +390,19 @@ class Master:
                                                ('dt', DecisionTree())], cfg_cache=self.cfg_cache) \
                         .fill_candidate(cs, self.ds)
 
-                if self.result_logger is not None:
-                    self.result_logger.new_structure(cs)
+                self.result_logger.new_structure(cs)
                 self.bandit_learner.iterations[-1].replace_proxy(cs)
 
                 self.incomplete_structures[cs.cid] = cs, int(cs.budget), 0
+
+                self.n_structures = self.n_structures + 1
             except KeyboardInterrupt:
                 raise
             except (BrokenPipeError, EOFError) as ex:
-                self.logger.fatal('Lost connection to SyncManager probably due to OOM. Aborting...: {}'.format(ex),
+                self.logger.fatal(f'Lost connection to SyncManager probably due to OOM. Aborting...: {ex}',
                                   exc_info=True)
                 self.abort = True
             except Exception as ex:
-                self.logger.fatal('Encountered unhandled exception {}. This should never happen!'.format(ex),
-                                  exc_info=True)
+                self.logger.fatal(f'Encountered unhandled exception {ex}. This should never happen!', exc_info=True)
             finally:
                 self.thread_cond.notify_all()
