@@ -1,14 +1,13 @@
-import abc
 import logging
 import math
+import operator
 import os
 import pickle
-import random
 import threading
 import timeit
 from abc import ABC
 from copy import deepcopy
-from typing import List, Optional, Tuple, Type, Dict
+from typing import List, Optional, Tuple, Type, Dict, Union, Any, Iterator
 
 import joblib
 import networkx as nx
@@ -22,35 +21,13 @@ from dswizard.components.data_preprocessing import DataPreprocessorChoice
 from dswizard.components.feature_preprocessing import FeaturePreprocessorChoice
 from dswizard.core.base_structure_generator import BaseStructureGenerator
 from dswizard.core.model import CandidateId, PartialConfig, StatusType, CandidateStructure, Dataset, Result, \
-    EvaluationJob
+    EvaluationJob, Runtime
 from dswizard.core.similaritystore import SimilarityStore
 from dswizard.core.worker import Worker
 from dswizard.pipeline.pipeline import FlexiblePipeline
 from dswizard.util import util
 
 
-# Ideas
-#   1. Bandit Learning
-#       - Iterate all pipelines with maximum depth d
-#       - Use MCTS to select next pipeline to tune
-#           - Use grammar to generate valid pipelines during simulation and expansion
-#           - Consider pipeline complexity to avoid over-fitting
-#           - Stop node selection to re-select a shorter pipeline
-#           - Prior for step performance via OpenML flows
-#           - Do not select next action indepently but consider previous steps. Algorithms are often selected in paris
-#               (see MLPlan)
-#       - Optimize hyperparameters via BO
-#           - Prior for hyperparameters via OpenML flows
-#
-#  General:
-#   - Multi-target optimization
-#       - Quality
-#       - Complexity
-#       - Training time
-#       - Hardware limitations
-#
-
-
 class Node:
     """
     A representation of a single board state.
@@ -58,21 +35,26 @@ class Node:
     Could be e.g. a chess or checkers board state.
     """
 
+    UNVISITED = 'Unvisited'
+    INCOMPLETE = 'Incomplete'
+    ROOT = 'Root'
+
     def __init__(self,
-                 id: int,
+                 node_id: int,
                  ds: Optional[Dataset],
                  component: Optional[Type[EstimatorComponent]],
                  pipeline_prefix: List[Tuple[str, EstimatorComponent]] = None,
-                 partial_config: PartialConfig = None
                  ):
-        self.id = id
+        self.id = node_id
         self.ds = ds
-        self.partial_config = partial_config
+        self.partial_config = None
+        self.runtime: Optional[Runtime] = None
 
         self.failure_message: Optional[str] = None
+        self.expanded = False
 
         if component is None:
-            self.label = 'ROOT'
+            self.label = Node.ROOT
             self.component = None
         else:
             self.component = component()
@@ -82,12 +64,12 @@ class Node:
             pipeline_prefix = []
         else:
             pipeline_prefix = deepcopy(pipeline_prefix)
-            # TODO check if also add if pipeline_prefix is None
-            pipeline_prefix.append((str(id), self.component))
+            pipeline_prefix.append((f'{len(pipeline_prefix)}:{self.label}', self.component))
         self.steps: List[Tuple[str, EstimatorComponent]] = pipeline_prefix
 
         self.visits = 0
         self.reward = 0
+        self.explanations: Dict[str, Dict[str, Any]] = {}
 
     def is_terminal(self):
         return self.component is not None and is_classifier(self.component)
@@ -96,6 +78,10 @@ class Node:
     def failed(self):
         return self.failure_message is not None
 
+    @property
+    def unvisited(self):
+        return self.failure_message is not None and self.failure_message == Node.UNVISITED
+
     # noinspection PyMethodMayBeStatic
     def available_actions(self,
                           include_preprocessing: bool = True,
@@ -107,29 +93,38 @@ class Node:
         if include_preprocessing:
             components.update(DataPreprocessorChoice().get_available_components(mf=mf))
             components.update(FeaturePreprocessorChoice().get_available_components(mf=mf))
+        if 'noop' in components:
+            del components['noop']
         return components
 
-    def enter(self):
+    def enter(self, cid: CandidateId):
         self.reward += util.worst_score(self.ds.metric)[-1]
+        self.explanations[cid.external_name]['selected'] = True
 
-    def exit(self):
+    def exit(self, cid: Optional[CandidateId] = None):
         self.reward -= util.worst_score(self.ds.metric)[-1]
+        if cid:
+            self.explanations[cid.external_name]['selected'] = False
 
     def update(self, reward: float) -> 'Node':
         self.visits += 1
         self.reward += reward
         return self
 
+    def record_explanation(self, cid: CandidateId, score: float, policy: Dict):
+        self.explanations[cid.external_name] = {
+            'failure_message': self.failure_message,
+            'score': score,
+            'selected': False,
+            'policy': policy
+        }
+
     def __hash__(self) -> int:
         return self.id
 
     def __eq__(self, node2: int) -> bool:
         return self.id == node2
 
-    @staticmethod
-    def dummy(component, perf: float) -> 'Node':
-        return Node(-1, None, component).update(perf)
-
 
 class Tree:
     ROOT: int = 0
@@ -137,7 +132,8 @@ class Tree:
     def __init__(self, ds: Dataset):
         self.G = nx.DiGraph()
         self.id_count = -1
-        self.add_node(ds=ds)
+        root = self.add_node(ds=ds)
+        root.runtime = Runtime(0, 0)
 
         self.coef_progressive_widening = 0.7
 
@@ -149,50 +145,66 @@ class Tree:
         new_id = self.id_count
         node = Node(new_id, ds, estimator, pipeline_prefix=parent_node.steps if parent_node is not None else None)
         if ds is None:
-            node.failure_message = 'Incomplete'
+            node.failure_message = Node.UNVISITED
+        if parent_node is None:
+            # ROOT node should always have at least 1 visit to be able to calculate UCT
+            node.visits += 1
 
-        self.G.add_node(new_id, value=node, label=node.label)
+        self.G.add_node(new_id, value=node)
         if parent_node is not None:
             self.G.add_edge(parent_node, new_id)
         return node
 
+    def inflate_node(self, estimator: Type[EstimatorComponent], parent_node: Node) -> Node:
+        children = self.get_children(parent_node.id, include_unvisited=True)
+        node = next(filter(lambda n: isinstance(n.component, estimator), children))
+        assert node.failure_message == Node.UNVISITED
+        node.failure_message = Node.INCOMPLETE
+        return node
+
     def get_node(self, node: int) -> Node:
         return self.G.nodes[node]['value']
 
-    def get_children(self, node: int) -> List[Node]:
-        return [self.G.nodes[n]['value'] for n in self.G.successors(node)]
+    def get_children(self, node: int, include_unvisited: bool = False) -> List[Node]:
+        nodes = [self.G.nodes[n]['value'] for n in self.G.successors(node)]
+        return list(filter(lambda n: include_unvisited or not n.unvisited, nodes))
 
-    def get_all_children(self, node: int) -> List[Node]:
-        return [self.G.nodes[n]['value'] for n in nx.dfs_tree(self.G, node).nodes.keys()]
+    def expand_node(self, node: Node):
+        for key, value in node.available_actions().items():
+            self.add_node(value, parent_node=node)
+        node.expanded = True
 
-    def fully_expanded(self, node: Node, global_max: int = 10):
+    def fully_expanded(self, node: Node, global_max: int = 7):
         # global_max is used as failsafe to prevent massive expansion of single node
 
         possible_children = len(node.available_actions())
         max_children = math.floor(math.pow(node.visits, self.coef_progressive_widening))
         max_children = min(max_children, possible_children, global_max)
-
-        current_children = len(list(self.G.successors(node.id)))
+        current_children = len(self.get_children(node.id))
 
         return possible_children == 0 or (current_children != 0 and current_children >= max_children)
 
+    def predecessors(self, node: Node) -> Iterator:
+        return self.G.predecessors(node.id)
+
     def plot(self, file: str):
-        # TODO plot can only be called once
-        for n, data in self.G.nodes(data=True):
+        h = self.G.copy()
+        unvisited_nodes = []
+        for n, data in h.nodes(data=True):
             node: Node = data['value']
-            if node.label.endswith('Component'):
-                data['label'] = node.label[:-9]
 
             score = node.reward / node.visits if node.visits > 0 else 0
-            data['label'] = '{} ({})\n{:.4f} / {}'.format(data['label'], n, score, node.visits)
+            data['label'] = f'{node.label} ({n})\n{score:.4f} / {node.visits}'
 
-            if node.failed:
+            if node.failure_message == Node.UNVISITED:
+                unvisited_nodes.append(n)
+            elif node.failed:
                 data['fillcolor'] = 'gray'
                 data['style'] = 'filled'
                 data['label'] += '\n' + node.failure_message
 
-        H = nx.nx_agraph.to_agraph(self.G)
-        H.draw(file, prog='dot')
+        h.remove_nodes_from(unvisited_nodes)
+        nx.nx_agraph.to_agraph(h).draw(file, prog='dot')
 
     def __contains__(self, node: int) -> bool:
         return node in self.G.nodes
@@ -200,43 +212,95 @@ class Tree:
 
 class Policy(ABC):
 
-    def __init__(self, logger: logging.Logger, exploration_weight: float = 2, wallclock_limit: float = None, **kwargs):
+    def __init__(self, logger: logging.Logger, exploration_weight: float = 1, wallclock_limit: float = None, **kwargs):
         self.logger = logger
         self.exploration_weight = exploration_weight
         self._exploration_weight = exploration_weight
         self.start = timeit.default_timer()
         self.wallclock_limit = wallclock_limit
 
-    @abc.abstractmethod
     def get_next_action(self,
-                        n: Node,
-                        current_children: List[Node],
+                        node: Node,
+                        children: List[Node],
+                        cid: CandidateId,
                         include_preprocessing: bool = True,
                         include_classifier: bool = True) -> Optional[Type[EstimatorComponent]]:
+        available_actions = node.available_actions()
+        actions = []
+        visited_children: List[Node] = list(filter(lambda _: not _.unvisited, children))
+
+        # Prefer classifier if no children have been visited yet
+        if include_classifier and len(visited_children) < 2:
+            actions = self._get_actions(node, visited_children, include_preprocessing=False, include_classifier=True)
+        if len(actions) == 0:
+            actions = self._get_actions(node, visited_children, include_preprocessing=include_preprocessing,
+                                        include_classifier=include_classifier)
+        if len(actions) == 0:
+            return None
+
+        estimated_performances = dict(zip(actions, self.estimate_performance(actions, node.ds)))
+        worst_performance = util.worst_score(node.ds.metric)[-1]
+        for n in children:
+            if not n.unvisited:
+                continue
+
+            name = n.component.component_name_
+            performance = worst_performance
+            if name in estimated_performances:
+                performance = -estimated_performances[name]
+
+            assert n.failure_message == Node.UNVISITED
+            n.visits += 1
+            n.reward += performance
+            n.failure_message = None
+
+            score, policy = self.uct(n, node, worst_score=worst_performance, decompose=True)
+            estimated_performances[name] = score
+
+            n.visits -= 1
+            n.reward -= performance
+            n.failure_message = Node.UNVISITED
+            n.record_explanation(cid, score, policy)
+
+        return available_actions[min(estimated_performances.items(), key=operator.itemgetter(1))[0]]
+
+    def estimate_performance(self, actions: List[str], ds: Dataset, depth: int = 1) -> np.ndarray:
         pass
 
-    def estimate_performance(self, actions: List[str], ds: Dataset, depth: int = 1):
-        return util.worst_score(ds.metric)[-1] * np.ones(len(actions))
-
-    def uct(self, n: Node, parent: Node, scale=2., force: bool = False) -> float:
+    def uct(self, n: Node, parent: Optional[Node], force: bool = False, worst_score: float = math.inf,
+            decompose: bool = False) -> Union[float, Tuple[float, Dict[str, float]]]:
         """Upper confidence bound for trees"""
-        if n.failed or n.visits == 0:
-            # Always ignore nodes without meta-features
-            return -math.inf
 
         # Always pick terminal node if forced
         if force and n.is_terminal():
-            return math.inf
+            return -math.inf
 
         if parent is None or parent.visits == 0:
             log_N_vertex = 0
         else:
             log_N_vertex = math.log(parent.visits)
 
-        exploitation = (-1 * n.reward) / n.visits  # HPO computes minimization problem. UCT selects maximum
-        exploration = math.sqrt(log_N_vertex / n.visits)
-        overfitting = (scale ** len(n.steps)) / (scale ** 10)
-        return (exploitation + self._exploration_weight * exploration) * (1 - overfitting)
+        if n.failed or n.visits == 0:
+            exploitation = worst_score
+            exploration = worst_score
+        else:
+            exploitation = n.reward / n.visits
+            exploration = -math.sqrt(log_N_vertex / n.visits)
+        score = exploitation + self._exploration_weight * exploration
+
+        overfitting = 1 - (2 ** len(n.steps)) / (2 ** 4)
+        adjusted_score = score * overfitting
+
+        if decompose:
+            return adjusted_score, {
+                'exploit': exploitation,
+                'explore': exploration,
+                'visits': n.visits - 1,  # node visits is always off by 1 to prevent division by 0
+                'overfit': overfitting,
+                'weight': self._exploration_weight
+            }
+        else:
+            return adjusted_score
 
     def select(self, node: Node, tree: Tree, force: bool = False) -> Tuple[Node, float]:
         """Select a child of node, balancing exploration & exploitation"""
@@ -247,23 +311,19 @@ class Policy(ABC):
                 children = terminal_children
             else:
                 children = current_children
-        elif not tree.fully_expanded(node):
-            available_actions = node.available_actions()
-            exhausted_actions = [type(n.component) for n in current_children]
-            actions = [key for key, value in available_actions.items() if value not in exhausted_actions]
-            perf = self.estimate_performance(actions, node.ds)
-
-            possible_children = [Node.dummy(available_actions[c], p) for c, p in zip(actions, perf)]
-            children = current_children + possible_children
         else:
             children = current_children
 
         if self.wallclock_limit is not None:
-            self._exploration_weight = max(0.0, self.exploration_weight * (
-                    math.exp((self.wallclock_limit + self.start - timeit.default_timer()) / self.wallclock_limit) - 1))
+            max_time = self.wallclock_limit
+            passed_time = timeit.default_timer() - self.start
+            self._exploration_weight = max(0.1, self.exploration_weight * (
+                    (math.exp((max_time - passed_time) / max_time) - math.exp(0)) / (math.exp(1) - math.exp(0))
+            ))
 
-        scores = [self.uct(n, node) for n in children]
-        idx = int(np.argmax(scores))
+        worst_score = util.worst_score(node.ds.metric)[-1]
+        scores = [self.uct(n, node, worst_score=worst_score) for n in children]
+        idx = int(np.argmin(scores))
 
         # Selected non-existing child. Mark as "failed" to force abortion of select
         if idx >= len(current_children):
@@ -271,22 +331,22 @@ class Policy(ABC):
 
         return children[idx], scores[idx]
 
+    @staticmethod
+    def _get_actions(node: Node,
+                     current_children: List[Node],
+                     include_preprocessing: bool,
+                     include_classifier: bool) -> List[str]:
+        available_actions = node.available_actions(include_preprocessing=include_preprocessing,
+                                                   include_classifier=include_classifier)
+        exhausted_actions = [type(n.component) for n in current_children]
+        actions = [key for key, value in available_actions.items() if value not in exhausted_actions]
+        return actions
+
 
 class RandomSelection(Policy):
 
-    def get_next_action(self,
-                        n: Node,
-                        current_children: List[Node],
-                        include_preprocessing: bool = True,
-                        include_classifier: bool = True) -> Optional[Type[EstimatorComponent]]:
-        actions = n.available_actions(include_preprocessing=include_preprocessing,
-                                      include_classifier=include_classifier).values()
-        exhausted_actions = [type(n.component) for n in current_children]
-        actions = [a for a in actions if a not in exhausted_actions]
-
-        if len(actions) == 0:
-            return None
-        return random.choice(actions)
+    def estimate_performance(self, actions: List[str], ds: Dataset, depth: int = 1):
+        return np.random.random(len(actions))
 
 
 class TransferLearning(Policy):
@@ -294,7 +354,7 @@ class TransferLearning(Policy):
     def __init__(self, logger: logging.Logger, model: str, **kwargs):
         super().__init__(logger, **kwargs)
 
-        logger.info('Loading transfer model from {}'.format(model))
+        logger.info(f'Loading transfer model from {model}')
         with open(model, 'rb') as f:
             mean, var = joblib.load(f)
         self.mean: Pipeline = mean
@@ -311,6 +371,10 @@ class TransferLearning(Policy):
         # remove land-marking mf
         X = np.delete(X, slice(42, 50), axis=1)
 
+        # meta-learning base does not contain ordinal_encoder yet. Retrain meta-learning base with new components
+        con = X[:, 42] == 'ordinal_encoder'
+        X[con, 42] = 'one_hot_encoding'
+
         mean = self.mean.predict(X)
         var = self.var.predict(X)
         var = np.maximum(var, 0.01 * np.ones(var.shape))
@@ -318,21 +382,6 @@ class TransferLearning(Policy):
         perf = np.random.multivariate_normal(mean, np.diag(var))
         return perf
 
-    def get_next_action(self,
-                        n: Node,
-                        current_children: List[Node],
-                        include_preprocessing: bool = True,
-                        include_classifier: bool = True) -> Optional[Type[EstimatorComponent]]:
-        available_actions = n.available_actions(include_preprocessing=include_preprocessing,
-                                                include_classifier=include_classifier)
-        exhausted_actions = [type(n.component) for n in current_children]
-        actions = [key for key, value in available_actions.items() if value not in exhausted_actions]
-        if len(actions) == 0:
-            return None
-
-        perf = self.estimate_performance(actions, n.ds)
-        return available_actions[actions[int(np.argmax(perf))]]
-
 
 class MCTS(BaseStructureGenerator):
     """Monte Carlo tree searcher. First rollout the tree then choose a move."""
@@ -352,9 +401,16 @@ class MCTS(BaseStructureGenerator):
         self.tree: Optional[Tree] = None
         self.store_ds = store_ds
         self.lock = threading.Lock()
+        self.cid_to_node = {}
+        self.wallclock_limit = wallclock_limit if epsilon_greedy else 1
+        self.start = timeit.default_timer()
 
         if policy is None:
-            policy = RandomSelection
+            if model is None:
+                policy = RandomSelection
+            else:
+                policy = TransferLearning
+
         policy_kwargs = {
             'model': model,
             'wallclock_limit': wallclock_limit if epsilon_greedy else None
@@ -364,7 +420,7 @@ class MCTS(BaseStructureGenerator):
             # noinspection PyUnresolvedReferences
             similarity_model = self.policy.mean
         except (KeyError, FileNotFoundError) as ex:
-            self.logger.warning('Failed to initialize Policy: {}. Fallback to RandomSelection.'.format(ex))
+            self.logger.warning(f'Failed to initialize Policy: {ex}. Fallback to RandomSelection.')
             self.policy = RandomSelection(self.logger)
             similarity_model = None
         except AttributeError:
@@ -378,23 +434,26 @@ class MCTS(BaseStructureGenerator):
         with self.lock:
             if self.tree is None:
                 self.tree = Tree(ds)
-                self.store.add(ds.meta_features)
+                self.store.add(ds.meta_features, data=Tree.ROOT)
+
+            self._record_explanations(cs.cid)
 
         if retries < 0:
             self.logger.warning('Retries exhausted. Aborting candidate sampling')
             return cs
 
         # traverse from root to a leaf node
-        path, expand = self._select(force=retries == 0)
+        path, expand = self._select(cid=cs.cid, force=retries == 0)
 
         result = None
         if expand or not path[-1].is_terminal():
             # expand last node in path if possible
             max_depths = 3
             max_failures = 3
-            timeout = None if cutoff is None else timeit.default_timer() + cutoff
+            timeout = None if cutoff is None or cutoff <= 0 else timeit.default_timer() + cutoff
             for i in range(1, max_depths + 1):
-                expansion, result, failures = self._expand(path, worker, cs.cid, timeout=timeout,
+                expansion, result, failures = self._expand(path, worker, cs.cid,
+                                                           timeout=timeout,
                                                            max_failures=max_failures,
                                                            include_preprocessing=i < max_depths)
                 max_failures -= failures
@@ -407,73 +466,91 @@ class MCTS(BaseStructureGenerator):
         else:
             self.logger.debug('Skipping MCTS expansion')
 
+        failed = False
         if len(path) == 1:
-            with self.lock:
-                for node in path:
-                    self.tree.get_node(node.id).exit()
-            self.logger.warning(
-                'Current path contains only ROOT node. Trying tree traversal {} more times...'.format(retries))
-            return self.fill_candidate(cs, ds, worker=worker, retries=retries - 1)
+            self.logger.warning(f'Current path contains only ROOT node. Trying tree traversal {retries} more times...')
+            failed = True
         if not path[-1].is_terminal():
+            self.logger.warning(f'Current path is not terminal. Trying tree traversal {retries} more times...')
+            failed = True
+        if failed:
             with self.lock:
                 for node in path:
-                    self.tree.get_node(node.id).exit()
-            self.logger.warning(
-                'Current path does not end in a classifier. Trying tree traversal {} more times...'.format(retries))
+                    self.tree.get_node(node.id).exit(cs.cid)
             return self.fill_candidate(cs, ds, worker=worker, retries=retries - 1)
 
         # A simulation is not necessary. Simulated results are already incorporated in the policy
 
         node = path[-1]
-        self.logger.debug('Sampled pipeline structure: {}'.format(node.steps))
+        self.cid_to_node[cs.cid] = node
+        self.logger.debug(f'Sampled pipeline structure: {node.steps}')
         pipeline = FlexiblePipeline(node.steps)
 
         cs.configspace = pipeline.configuration_space
         cs.pipeline = pipeline
-        cs.cfg_keys = [n.partial_config.cfg_key for n in path if n.partial_config is not None]
+        cs.cfg_keys = [n.partial_config.cfg_key for n in path[1:]]
 
         # If no simulation was necessary, add the default configuration as first result
         if result is not None and result.structure_loss is not None and \
                 result.structure_loss < util.worst_score(ds.metric)[-1]:
             cs.add_result(result)
+
+        assert len(cs.steps) == len(cs.cfg_keys)
         return cs
 
-    def _select(self, force: bool = False) -> Tuple[List[Node], bool]:
+    def _select(self, cid: CandidateId, force: bool = False) -> Tuple[List[Node], bool]:
         """Find an unexplored descendent of ROOT"""
 
         path: List[Node] = []
         node = self.tree.get_node(Tree.ROOT)
         score = -math.inf
+        epsilon = (timeit.default_timer() - self.start) / self.wallclock_limit
+
+        # Failsafe mechanism to enforce structure selection
+        if force or np.random.random() < epsilon / 2:
+            nodes = [self.tree.get_node(n) for n in nx.dfs_tree(self.tree.G, self.tree.ROOT)]
+            terminal_nodes = [n for n in nodes if n.is_terminal() and not n.failed]
+            probs = np.array([n.reward / n.visits for n in terminal_nodes]) * -1
+            node = np.random.choice(terminal_nodes, 1, p=probs / np.sum(probs))[0]
+
+            path = nx.shortest_path(self.tree.G, source=self.tree.ROOT, target=node.id)
+            # noinspection PyTypeChecker
+            path[0] = self.tree.get_node(path[0])
+            # noinspection PyTypeChecker
+            path[-1] = self.tree.get_node(path[-1])
+
+            with self.lock:
+                for node in path:
+                    node.enter(cid)
+            return path, False
+
         while True:
-            self.logger.debug('\tSelecting {}'.format(node.label))
+            self.logger.debug(f'\tSelecting {node.label}')
             if node.failed:
                 self.logger.warning(
                     'Selected node has no dataset, meta-features or partial configurations. This should not happen')
 
             with self.lock:
-                node.enter()
+                node.enter(cid)
                 path.append(node)
 
-                # Failsafe mechanism to enforce structure selection
-                if force and node.is_terminal():
-                    return path, False
-
+                # We never expand if force is given
                 fully_expanded = self.tree.fully_expanded(node)
                 if not fully_expanded:
                     return path, True
 
                 # node is leaf
                 if len(self.tree.get_children(node.id)) == 0:
-                    return path, True
+                    return path, node.is_terminal()
 
-                candidate, candidate_score = self.policy.select(node, self.tree, force=force)  # descend a layer deeper
+                candidate, candidate_score = self.policy.select(node, self.tree)
                 # Current node is better than children or selected node failed
                 if candidate.failed:
                     return path, not node.is_terminal()
-                elif candidate_score <= score and node.is_terminal():
+                elif candidate_score >= score and node.is_terminal():
                     return path, False
-                elif candidate_score <= score and not fully_expanded:
-                    return path, True
+                elif candidate_score >= score and not fully_expanded:
+                    return path, not node.is_terminal()
                 else:
                     node, score = candidate, candidate_score
 
@@ -496,40 +573,47 @@ class MCTS(BaseStructureGenerator):
                     self.logger.warning('Aborting expansion due to timeout')
                     return None, None, max_failures
 
+                if not node.expanded:
+                    self.tree.expand_node(node)
                 n_children = len(self.tree.get_children(node.id))
                 if n_children >= n_actions:
                     break
 
-                current_children = self.tree.get_children(node.id)
-                action = self.policy.get_next_action(node, current_children,
-                                                     include_preprocessing=include_preprocessing)
+                children = self.tree.get_children(node.id, include_unvisited=True)
+                action = self.policy.get_next_action(node, children, cid, include_preprocessing=include_preprocessing)
                 if action is None:
                     return None, None, failure_count
                 if failure_count >= max_failures:
-                    self.logger.warning('Aborting expansion due to {} failed expansions'.format(failure_count))
+                    self.logger.warning(f'Aborting expansion due to {failure_count} failed expansions')
                     return None, None, failure_count
 
-                component = action()
-
-                self.logger.debug(
-                    '\tExpanding with {}. Option {}/{}'.format(component.name(), n_children + 1, n_actions))
-                new_node = self.tree.add_node(estimator=action, parent_node=node)
+                new_node = self.tree.inflate_node(estimator=action, parent_node=node)
+                component = new_node.component
+                self.logger.debug(f'\tExpanding with {component.name()}. Option {n_children + 1}/{n_actions}')
 
             ds = node.ds
             config, key = self.cfg_cache.sample_configuration(
+                cid=cid.with_config(0),
+                name=new_node.steps[-1][0],
                 configspace=component.get_hyperparameter_search_space(),
                 mf=ds.meta_features,
                 default=True)
 
-            job = EvaluationJob(ds, cid.with_config(-new_node.id), cs=component, cutoff=self.cutoff, config=config,
-                                cfg_keys=[key])
+            job = EvaluationJob(ds, cid.with_config(f'{len(node.steps)}_{component.name(short=True)}'),
+                                cs=component, cutoff=self.cutoff, config=config, cfg_keys=[key])
             result = worker.start_transform_dataset(job)
 
-            if result.status.value == StatusType.SUCCESS.value:
+            # TODO remove again
+            if len(node.steps) < 4 and \
+                    result.status.value == StatusType.SUCCESS.value:
                 ds = Dataset(result.transformed_X, ds.y, ds.metric, ds.cutoff)
                 new_node.partial_config = PartialConfig(key, config, str(new_node.id), ds.meta_features)
                 new_node.ds = ds
 
+                # Add trainings time of all previous nodes to new node
+                result.runtime.training_time += sum([n.runtime.training_time for n in nodes])
+                new_node.runtime = result.runtime
+
                 if ds.meta_features is None:
                     result.status = StatusType.CRASHED
                     result.structure_loss = util.worst_score(ds.metric)[-1]
@@ -538,32 +622,29 @@ class MCTS(BaseStructureGenerator):
                     failure_count += 1
                 else:
                     # Check if any node in the tree is similar to the new dataset
-                    distance, idx = self.store.get_similar(ds.meta_features)
+                    distance, _, idx = self.store.get_similar(ds.meta_features)
                     if np.allclose(node.ds.meta_features, ds.meta_features, equal_nan=True):
-                        self.logger.debug('\t{} did not modify dataset'.format(component.name()))
+                        self.logger.debug(f'\t{component.name()} did not modify dataset')
                         result.status = StatusType.INEFFECTIVE
                         result.structure_loss = util.worst_score(ds.metric)[-1]
                         new_node.failure_message = 'Ineffective'
-                    elif distance[0][0] <= max_distance:
+                    elif distance <= max_distance:
                         # TODO: currently always the existing node is selected. This node could represent simpler model
-                        self.logger.debug('\t{} produced a dataset similar to {}'.format(component.name(), idx[0][0]))
+                        self.logger.debug(f'\t{component.name()} produced a dataset similar to {idx}')
                         result.status = StatusType.DUPLICATE
                         result.structure_loss = util.worst_score(ds.metric)[-1]
-                        new_node.failure_message = 'Duplicate {}'.format(idx[0][0])
+                        new_node.failure_message = f'Duplicate {idx}'
                     else:
-                        self.store.add(ds.meta_features)
-                        node.enter()
-
-                        # Hacky solution. If result loss is set, 'Incomplete' message is removed later
-                        if result.structure_loss is None:
-                            new_node.failure_message = None
+                        self.store.add(ds.meta_features, data=new_node.id)
+                        # Enter node as enter was not called during tree traversal yet
+                        new_node.enter(cid)
+                        new_node.failure_message = None
 
                         if self.store_ds:
-                            with open(os.path.join(self.workdir, '{}.pkl'.format(new_node.id)), 'wb') as f:
+                            with open(os.path.join(self.workdir, f'{new_node.id}.pkl'), 'wb') as f:
                                 pickle.dump(ds, f)
             else:
-                self.logger.debug(
-                    '\t{} failed with default hyperparamter: {}'.format(component.name(), result.status))
+                self.logger.debug(f'\t{component.name()} failed with default hyperparamter: {result.status}')
                 result.structure_loss = util.worst_score(ds.metric)[-1]
                 if result.status == StatusType.TIMEOUT:
                     new_node.failure_message = 'Timeout'
@@ -572,12 +653,13 @@ class MCTS(BaseStructureGenerator):
 
             if result.structure_loss is not None:
                 n_children += 1
-                self._backpropagate([key for key, values in new_node.steps], result.structure_loss)
+                self._backpropagate(new_node, result.structure_loss)
                 if result.structure_loss < util.worst_score(ds.metric)[-1]:
-                    new_node.failure_message = None
-                    result.partial_configs = [n.partial_config for n in nodes if n.partial_config is not None]
+                    result.partial_configs = [n.partial_config for n in nodes[1:]]
                     result.partial_configs.append(new_node.partial_config)
-                    result.config = FlexiblePipeline(new_node.steps).configuration_space.get_default_configuration()
+                    config = FlexiblePipeline(new_node.steps).configuration_space.get_default_configuration()
+                    config.origin = 'Default'
+                    result.config = config
 
                     job.result = result
                     self.cfg_cache.register_result(job)
@@ -589,31 +671,74 @@ class MCTS(BaseStructureGenerator):
                 return new_node, result, failure_count
         return None, None, failure_count
 
+    def _record_explanations(self, cid: CandidateId):
+        root = self.tree.get_node(self.tree.ROOT)
+        worst_score = util.worst_score(root.ds.metric)[-1]
+        for node_id in self.tree.G.nodes:
+            node = self.tree.get_node(node_id)
+            if node.visits == 0:
+                # Exclude unvisited nodes from tree traversal
+                continue
+            try:
+                parent = next(self.tree.G.predecessors(node_id))
+            except StopIteration:
+                parent = None
+            score, policy = self.policy.uct(node, parent, worst_score=worst_score, decompose=True)
+            node.record_explanation(cid, score, policy)
+
     def register_result(self, candidate: CandidateStructure, result: Result, update_model: bool = True,
                         **kwargs) -> None:
         reward = result.structure_loss
 
         if reward is None or not np.isfinite(reward):
-            reward = 1
+            reward = util.worst_score(self.tree.get_node(self.tree.ROOT).ds.metric)[-1]
 
         try:
-            self._backpropagate(candidate.pipeline.steps_.keys(), reward, exit=True)
-        except (IndexError, ValueError, AttributeError) as ex:
-            self.logger.warning('Unable to backpropagate results: {}'.format(ex))
+            node = self.cid_to_node[candidate.cid]
+            self._backpropagate(node, reward, exit=True)
+        except (IndexError, ValueError, AttributeError, KeyError) as ex:
+            self.logger.warning(f'Unable to backpropagate results: {ex}')
 
     # noinspection PyMethodMayBeStatic
-    def _backpropagate(self, path: List[str], reward: float, exit: bool = False) -> None:
+    def _backpropagate(self, node: Node, reward: float, exit: bool = False) -> None:
         """Send the reward back up to the ancestors of the leaf"""
-        # Also update root node
-
-        ls = list(path)
-        ls.append('0')
         with self.lock:
-            for node_id in ls:
-                node = self.tree.get_node(int(node_id))
-                node.update(reward)
+            nodes = [node] + list(self.tree.predecessors(node)) + [self.tree.get_node(self.tree.ROOT)]
+
+            for pred in nodes:
+                pred.update(reward)
                 if exit:
-                    node.exit()
+                    pred.exit()
+
+    def explain(self) -> Dict[str, Any]:
+        with self.lock:
+            nodes = {}
+            edges = nx.dfs_successors(self.tree.G, source=Tree.ROOT)
+
+            for node_id in self.tree.G.nodes:
+                node = self.tree.get_node(node_id)
+                nodes[node_id] = {
+                    'label': node.label,
+                    'details': node.explanations
+                }
+
+            def transform_node(node_id: int):
+                children = []
+                element = {
+                    'id': str(node_id),
+                    **nodes[node_id]
+                }
+                try:
+                    for child in edges[node_id]:
+                        children.append(transform_node(child))
+                except KeyError:
+                    pass
+                if len(children) > 0:
+                    element['children'] = children
+                return element
+
+            hierarchy = transform_node(Tree.ROOT)
+            return hierarchy
 
     def shutdown(self):
         if self.tree is None:
@@ -623,4 +748,4 @@ class MCTS(BaseStructureGenerator):
             self.tree.plot(os.path.join(self.workdir, 'search_graph.pdf'))
         except ImportError as ex:
             self.logger.warning("Saving search graph is not possible. Please ensure that visualization "
-                                "is correctly setup: {}".format(ex))
+                                f"is correctly setup: {ex}")
